!!$
!!$=head1 NAME
!!$
!!$ Phase - Phaseshift migration (possibly split-step)
!!$
!!$=head1 SYNOPSIS
!!$
!!$Phase par=parfile.P D=input.H  R=output.H S=slow.H
!!$
!!$=head1 COORDINATE SYSTEM FOR DOWNWARD CONTINUATION DESCRIPTION PARAMETERS
!!$
!!$=over 4
!!$
!!$
!!$=item amx_n (integer)
!!$
!!$  Number of CMP in X (axis 1 of D,R and S)
!!$
!!$=item amx_d (integer) 
!!$
!!$  Sampling of CMP in X
!!$
!!$=item amx_o (integer)
!!$
!!$  First CMP  X location 
!!$
!!$=item amy__n (integer)
!!$
!!$  Number of CMP in X (axis 2 of D,R and S)
!!$
!!$=item amy_d (integer) 
!!$
!!$  Sampling of CMP in Y
!!$
!!$=item amy_o (integer)
!!$
!!$  First CMP  Y location 
!!$
!!$=item ahx__n (integer)
!!$
!!$  Number of half offsets in X (axis 3 of D and possibly R  )
!!$
!!$=item ahx_d (integer) 
!!$
!!$  Sampling of half offsets in X
!!$
!!$=item ahx_o (integer)
!!$
!!$  First half offset  X location 
!!$
!!$=item ahy__n (integer)
!!$
!!$  Number of half offsets in Y (axis 4 of D and possibly R)
!!$
!!$=item ahy_d (integer) 
!!$
!!$  Sampling of half offset in Y
!!$
!!$=item ahy_o (integer)
!!$
!!$  First CMP  Y location 
!!$
!!$
!!$=back
!!$
!!$
!!$=head1 R/S  DESCRIPTION PARAMETERS
!!$
!!$=over 4
!!$
!!$
!!$=item az__n (integer) (axis 3 of S, axis 5 of R)
!!$
!!$  Number of depth steps
!!$
!!$=item az__d (integer)
!!$
!!$  Sampling of depths
!!$
!!$=item az__o (integer)
!!$
!!$  First depth of model
!!$
!!$=item image_real (logcal)
!!$
!!$  Whether or not make the image real
!!$
!!$=back
!!$
!!$
!!$=head1 DATA DESCRIPTION PARAMETERS
!!$
!!$=over 4
!!$
!!$=item aw__n (integer) (axis 5 of D)
!!$
!!$  Number of frequencies
!!$
!!$=item aw__d (integer)
!!$
!!$  Sampling of frequencies
!!$
!!$=item aw__o (integer)
!!$
!!$  First frequency of data
!!$
!!$=back
!!$
!!$=head1 SLOWNESS PARAMETERS
!!$
!!$=over 4
!!$
!!$=item  velocity  (logical)
!!$
!!$ [.false.] whether or not S is velocity
!!$
!!$=item  nref  (integer)
!!$
!!$ [1] number of velocities for each depth step 
!!$
!!$=back
!!$
!!$=head1 PH GATHERS
!!$
!!$=over 4
!!$ 
!!$=item pCIG (logical)
!!$
!!$ [.false.] whether (.true.) or not (.false.) to create PH gathers
!!$
!!$=item apx__n (integer) 
!!$
!!$  Number of PH in X (axis 3 of R)
!!$
!!$=item apx_d (integer) 
!!$
!!$  Sampling of PH in X
!!$
!!$=item apx_o (integer)
!!$
!!$  First PH X location 
!!$
!!$=item apy__n (integer)
!!$
!!$  Number of PH in Y (axis 4 of R)
!!$
!!$=item apy_d (integer) 
!!$
!!$  Sampling of PH in Y
!!$
!!$=item apy_o (integer)
!!$
!!$  First PH  Y location 
!!$
!!$=item pcigeps (float)
!!$
!!$  [.1] Epsilon parameter for Phgather construction
!!$
!!$
!!$=back
!!$
!!$=head1 MEMORY MANAGEMET
!!$
!!$=over 4
!!$
!!$=item nzs (integer)
!!$
!!$Number of Z steps to do in memory
!!$
!!$=item nws (integer)
!!$
!!$Number of W to hold in memory
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Perform split step phase shift migration/modeling
!!$
!!$
!!$
!!$
!!$=head1 COMMENTS
!!$
!!$=over 4
!!$
!!$=item  anx_n,ahx_d,ahx_o and axis 1 of D and R  must be the same
!!$
!!$=item  any_n,ahy_d,ahy_o and axis 2 of D and R  must be the same
!!$
!!$=item Slowness must contain entire  span of X and Y
!!$
!!$=item  az__n,az__d,az__o , axis 3 of S and axis 5 of R  must be the same
!!$
!!$=item When doing ph gathers apx_n = ahx_n = n3 of D
!!$
!!$=item When doing ph gathers apy_n = ahy_n = n4 of D
!!$
!!$=item When doing offset gathers  ahx_n,ahx_d,ahx_o = axis 3 of D and R
!!$
!!$=item When doing offset gathers  ahy_n,ahy_d,ahy_o = axis 4 of D and R
!!$
!!$=item nzs must be a multiple of az__n
!!$
!!$=item nws*nmpi_threads*nomp_threads must be a multiple of aw__n 
!!$
!!$=head1 CATEGORY
!!$
!!$B<seis/image>
!!$
!!$=cut
!!$
!!$
!!$Authors:
!!$
!!$ Paul Sava
!!$
!!$ Bob Clapp
!!$
!!$

program Phase   
  use wei_mig_mid_basic
  use wei_mig_var
  use wei_cig
  use wei_kxmig
  use wei_sll
  use wei_ssf
  use wei_wem
  use sep

  implicit none
  integer :: stat 

  integer :: i,n,ierr,iverb,nref
  logical :: verb
  integer :: nfk
  logical :: pCIG
  character(len=128) :: order


  !----------------------------------------------------------------
#ifdef SEP_MPI
  call MPI_INIT(stat)
  call  set_no_putch()
#endif
  call sep_init()
call sep_add_doc_line("NAME")
call sep_add_doc_line("     Phase - Phaseshift migration (possibly split-step)")
call sep_add_doc_line("")
call sep_add_doc_line("SYNOPSIS")
call sep_add_doc_line("    Phase par=parfile.P D=input.H R=output.H S=slow.H")
call sep_add_doc_line("")
call sep_add_doc_line("COORDINATE SYSTEM FOR DOWNWARD CONTINUATION DESCRIPTION PARAMETERS")
call sep_add_doc_line("    amx_n (integer)")
call sep_add_doc_line("          Number of CMP in X (axis 1 of D,R and S)")
call sep_add_doc_line("")
call sep_add_doc_line("    amx_d (integer)")
call sep_add_doc_line("          Sampling of CMP in X")
call sep_add_doc_line("")
call sep_add_doc_line("    amx_o (integer)")
call sep_add_doc_line("          First CMP  X location ")
call sep_add_doc_line("")
call sep_add_doc_line("    amy__n (integer)")
call sep_add_doc_line("          Number of CMP in X (axis 2 of D,R and S)")
call sep_add_doc_line("")
call sep_add_doc_line("    amy_d (integer)")
call sep_add_doc_line("          Sampling of CMP in Y")
call sep_add_doc_line("")
call sep_add_doc_line("    amy_o (integer)")
call sep_add_doc_line("          First CMP  Y location ")
call sep_add_doc_line("")
call sep_add_doc_line("    ahx__n (integer)")
call sep_add_doc_line("          Number of half offsets in X (axis 3 of D and possibly R  )")
call sep_add_doc_line("")
call sep_add_doc_line("    ahx_d (integer)")
call sep_add_doc_line("          Sampling of half offsets in X")
call sep_add_doc_line("")
call sep_add_doc_line("    ahx_o (integer)")
call sep_add_doc_line("          First half offset  X location ")
call sep_add_doc_line("")
call sep_add_doc_line("    ahy__n (integer)")
call sep_add_doc_line("          Number of half offsets in Y (axis 4 of D and possibly R)")
call sep_add_doc_line("")
call sep_add_doc_line("    ahy_d (integer)")
call sep_add_doc_line("          Sampling of half offset in Y")
call sep_add_doc_line("")
call sep_add_doc_line("    ahy_o (integer)")
call sep_add_doc_line("          First CMP  Y location ")
call sep_add_doc_line("")
call sep_add_doc_line("R/S  DESCRIPTION PARAMETERS")
call sep_add_doc_line("    az__n (integer) (axis 3 of S, axis 5 of R)")
call sep_add_doc_line("          Number of depth steps")
call sep_add_doc_line("")
call sep_add_doc_line("    az__d (integer)")
call sep_add_doc_line("          Sampling of depths")
call sep_add_doc_line("")
call sep_add_doc_line("    az__o (integer)")
call sep_add_doc_line("          First depth of model")
call sep_add_doc_line("")
call sep_add_doc_line("    image_real (logcal)")
call sep_add_doc_line("          Whether or not make the image real")
call sep_add_doc_line("")
call sep_add_doc_line("DATA DESCRIPTION PARAMETERS")
call sep_add_doc_line("    aw__n (integer) (axis 5 of D)")
call sep_add_doc_line("          Number of frequencies")
call sep_add_doc_line("")
call sep_add_doc_line("    aw__d (integer)")
call sep_add_doc_line("          Sampling of frequencies")
call sep_add_doc_line("")
call sep_add_doc_line("    aw__o (integer)")
call sep_add_doc_line("          First frequency of data")
call sep_add_doc_line("")
call sep_add_doc_line("SLOWNESS PARAMETERS")
call sep_add_doc_line("    velocity (logical)")
call sep_add_doc_line("         [.false.] whether or not S is velocity")
call sep_add_doc_line("")
call sep_add_doc_line("    nref (integer)")
call sep_add_doc_line("         [1] number of velocities for each depth step ")
call sep_add_doc_line("")
call sep_add_doc_line("PH GATHERS")
call sep_add_doc_line("    pCIG (logical)")
call sep_add_doc_line("         [.false.] whether (.true.) or not (.false.) to create PH gathers")
call sep_add_doc_line("")
call sep_add_doc_line("    apx__n (integer)")
call sep_add_doc_line("          Number of PH in X (axis 3 of R)")
call sep_add_doc_line("")
call sep_add_doc_line("    apx_d (integer)")
call sep_add_doc_line("          Sampling of PH in X")
call sep_add_doc_line("")
call sep_add_doc_line("    apx_o (integer)")
call sep_add_doc_line("          First PH X location ")
call sep_add_doc_line("")
call sep_add_doc_line("    apy__n (integer)")
call sep_add_doc_line("          Number of PH in Y (axis 4 of R)")
call sep_add_doc_line("")
call sep_add_doc_line("    apy_d (integer)")
call sep_add_doc_line("          Sampling of PH in Y")
call sep_add_doc_line("")
call sep_add_doc_line("    apy_o (integer)")
call sep_add_doc_line("          First PH  Y location ")
call sep_add_doc_line("")
call sep_add_doc_line("    pcigeps (float)")
call sep_add_doc_line("          [.1] Epsilon parameter for Phgather construction")
call sep_add_doc_line("")
call sep_add_doc_line("MEMORY MANAGEMET")
call sep_add_doc_line("    nzs (integer)")
call sep_add_doc_line("        Number of Z steps to do in memory")
call sep_add_doc_line("")
call sep_add_doc_line("    nws (integer)")
call sep_add_doc_line("        Number of W to hold in memory")
call sep_add_doc_line("")
call sep_add_doc_line("DESCRIPTION")
call sep_add_doc_line("     Perform split step phase shift migration/modeling")
call sep_add_doc_line("")
call sep_add_doc_line("COMMENTS")
call sep_add_doc_line("    anx_n,ahx_d,ahx_o and axis 1 of D and R must be the same")
call sep_add_doc_line("    any_n,ahy_d,ahy_o and axis 2 of D and R must be the same")
call sep_add_doc_line("    Slowness must contain entire span of X and Y")
call sep_add_doc_line("    az__n,az__d,az__o , axis 3 of S and axis 5 of R must be the same")
call sep_add_doc_line("    When doing ph gathers apx_n = ahx_n = n3 of D")
call sep_add_doc_line("    When doing ph gathers apy_n = ahy_n = n4 of D")
call sep_add_doc_line("    When doing offset gathers ahx_n,ahx_d,ahx_o = axis 3 of D and R")
call sep_add_doc_line("    When doing offset gathers ahy_n,ahy_d,ahy_o = axis 4 of D and R")
call sep_add_doc_line("    nzs must be a multiple of az__n")
call sep_add_doc_line("    nws*nmpi_threads*nomp_threads must be a multiple of aw__n")
call sep_add_doc_line("")
call sep_add_doc_line("CATEGORY")
call sep_add_doc_line("        seis/image")
call sep_add_doc_line("")

  !----------------------------------------------------------------
#ifdef SEP_MPI
  call MPI_COMM_SIZE(MPI_COMM_WORLD, n,ierr)
  call MPI_COMM_RANK(MPI_COMM_WORLD, i,ierr)
  if(i==0) then
     call from_param("verb",verb,.false.)
     iverb=0;if(verb) iverb=1
     call MPI_SEP_SEND_ARGS(n,10,iverb)
  else
     call MPI_SEP_RECEIVE_ARGS()
  end if
#endif


  call from_param("nref",nref,1)
  call from_param("pCIG",pCIG,.false.)
  call from_param("order",order,"OF")


  !----------------------------------------------------------------
  run%operator=".R_.D"
  !----------------------------------------------------------------

  if(.not. pCIG) then
    call weimig_init(   SLin=weisll_init	&
    ,                WCin=weikxmign_init	&
    ,                   FKin=weiwem_init	&
    ,                   FXin=weissf_init	&
    ,                   IGin=weihcig_init	)
   else
    call weimig_init(   SLin=weisll_init	&
    ,                WCin=weikxmign_init	&
    ,                   FKin=weiwem_init	&
    ,                   FXin=weissf_init	&
    ,                   IGin=weipcig_init	)
   end if
   if(.not. pCIG) then
     if(order(1:3)== "VAR") then
       stat = weimigvar(down%adj,down%add,down%R,down%D &
       ,              SLop=weisll	&
       ,           WCop=weikxmign_mid	&
       ,              FKop=weiwem	&
       ,              FXop=weissf	&
       ,              IGop=weihcig)
     else !1 velocity pCIG 
       stat = weimig(down%adj,down%add,down%R,down%D &
       ,              SLop=weisll	&
       ,           WCop=weikxmign_mid	&
       ,              FKop=weiwem	&
       ,              FXop=weissf	&
       ,              IGop=weihcig)
     end if
   else !1 velocity pCIG 
     if(order(1:3)== "VAR") then
       stat = weimigvar(down%adj,down%add,down%R,down%D &
       ,              SLop=weisll	&
       ,           WCop=weikxmign_mid	&
       ,              FKop=weiwem	&
       ,              FXop=weissf	&
       ,              IGop=weipcig)
     else !1 velocity pCIG 
       stat = weimigvar(down%adj,down%add,down%R,down%D &
       ,              SLop=weisll	&
       ,           WCop=weikxmign_mid	&
       ,              FKop=weiwem	&
       ,              FXop=weissf	&
       ,              IGop=weipcig)
     end if
endif

#ifdef SEP_MPI
  call MPI_FINALIZE(ierr) 
#endif
  call sep_close()
  !----------------------------------------------------------------
end program Phase
